@@ -1,14 +1,14 @@
 /*-------------------------------------------------------------------------
  *
- * junk.c
+ * execJunk.c
  *	  Junk attribute support stuff....
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/executor/execJunk.c,v 1.32 2002/09/04 20:31:17 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/executor/execJunk.c,v 1.46 2004/12/31 21:59:45 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -26,7 +26,7 @@
  * An attribute of a tuple living inside the executor, can be
  * either a normal attribute or a "junk" attribute. "junk" attributes
  * never make it out of the executor, i.e. they are never printed,
- * returned or stored in disk. Their only purpose in life is to
+ * returned or stored on disk. Their only purpose in life is to
  * store some information useful only to the executor, mainly the values
  * of some system attributes like "ctid" or rule locks.
  *
@@ -50,140 +50,42 @@
  *-------------------------------------------------------------------------
  */
 
-/*-------------------------------------------------------------------------
+/*
  * ExecInitJunkFilter
  *
  * Initialize the Junk filter.
  *
- * The initial targetlist and associated tuple descriptor are passed in.
+ * The source targetlist is passed in.  The output tuple descriptor is
+ * built from the non-junk tlist entries, plus the passed specification
+ * of whether to include room for an OID or not.
  * An optional resultSlot can be passed as well.
- *-------------------------------------------------------------------------
  */
 JunkFilter *
-ExecInitJunkFilter(List *targetList, TupleDesc tupType,
-				   TupleTableSlot *slot)
+ExecInitJunkFilter(List *targetList, bool hasoid, TupleTableSlot *slot)
 {
-	MemoryContext oldContext;
-	MemoryContext junkContext;
 	JunkFilter *junkfilter;
-	List	   *cleanTargetList;
-	int			len,
-				cleanLength;
 	TupleDesc	cleanTupType;
-	List	   *t;
-	TargetEntry *tle;
-	Resdom	   *resdom,
-			   *cleanResdom;
-	bool		resjunk;
-	AttrNumber	cleanResno;
+	int			cleanLength;
 	AttrNumber *cleanMap;
-	Node	   *expr;
+	ListCell   *t;
+	AttrNumber	cleanResno;
 
 	/*
-	 * Make a memory context that will hold the JunkFilter as well as all
-	 * the subsidiary structures we are about to create.  We use smaller-
-	 * than-default sizing parameters since we don't expect a very large
-	 * volume of stuff here.
+	 * Compute the tuple descriptor for the cleaned tuple.
 	 */
-	junkContext = AllocSetContextCreate(CurrentMemoryContext,
-										"JunkFilterContext",
-										1024,
-										1024,
-										ALLOCSET_DEFAULT_MAXSIZE);
-	oldContext = MemoryContextSwitchTo(junkContext);
+	cleanTupType = ExecCleanTypeFromTL(targetList, hasoid);
 
 	/*
-	 * First find the "clean" target list, i.e. all the entries in the
-	 * original target list which have a false 'resjunk' NOTE: make copy
-	 * of the Resdom nodes, because we have to change the 'resno's...
-	 */
-	cleanTargetList = NIL;
-	cleanResno = 1;
-
-	foreach(t, targetList)
-	{
-		TargetEntry *rtarget = lfirst(t);
-
-		if (rtarget->resdom != NULL)
-		{
-			resdom = rtarget->resdom;
-			expr = rtarget->expr;
-			resjunk = resdom->resjunk;
-			if (!resjunk)
-			{
-				/*
-				 * make a copy of the resdom node, changing its resno.
-				 */
-				cleanResdom = (Resdom *) copyObject(resdom);
-				cleanResdom->resno = cleanResno;
-				cleanResno++;
-
-				/*
-				 * create a new target list entry
-				 */
-				tle = makeTargetEntry(cleanResdom, expr);
-				cleanTargetList = lappend(cleanTargetList, tle);
-			}
-		}
-		else
-		{
-#ifdef SETS_FIXED
-			List	   *fjListP;
-			Fjoin	   *cleanFjoin;
-			List	   *cleanFjList;
-			List	   *fjList = lfirst(t);
-			Fjoin	   *fjNode = (Fjoin *) tl_node(fjList);
-
-			cleanFjoin = (Fjoin) copyObject((Node) fjNode);
-			cleanFjList = makeList1(cleanFjoin);
-
-			resdom = (Resdom) lfirst(get_fj_innerNode(fjNode));
-			expr = lsecond(get_fj_innerNode(fjNode));
-			cleanResdom = (Resdom) copyObject((Node) resdom);
-			set_resno(cleanResdom, cleanResno);
-			cleanResno++;
-			tle = (List) makeTargetEntry(cleanResdom, (Node *) expr);
-			set_fj_innerNode(cleanFjoin, tle);
-
-			foreach(fjListP, lnext(fjList))
-			{
-				TargetEntry *tle = lfirst(fjListP);
-
-				resdom = tle->resdom;
-				expr = tle->expr;
-				cleanResdom = (Resdom *) copyObject((Node) resdom);
-				cleanResno++;
-				cleanResdom->Resno = cleanResno;
-
-				/*
-				 * create a new target list entry
-				 */
-				tle = (List) makeTargetEntry(cleanResdom, (Node *) expr);
-				cleanFjList = lappend(cleanFjList, tle);
-			}
-			lappend(cleanTargetList, cleanFjList);
-#endif
-		}
-	}
-
-	/*
-	 * Now calculate the tuple type for the cleaned tuple (we were already
-	 * given the type for the original targetlist).
-	 */
-	cleanTupType = ExecTypeFromTL(cleanTargetList, tupType->tdhasoid);
-
-	len = ExecTargetListLength(targetList);
-	cleanLength = ExecTargetListLength(cleanTargetList);
-
-	/*
-	 * Now calculate the "map" between the original tuple's attributes and
+	 * Now calculate the mapping between the original tuple's attributes and
 	 * the "clean" tuple's attributes.
 	 *
 	 * The "map" is an array of "cleanLength" attribute numbers, i.e. one
 	 * entry for every attribute of the "clean" tuple. The value of this
 	 * entry is the attribute number of the corresponding attribute of the
-	 * "original" tuple.
+	 * "original" tuple.  (Zero indicates a NULL output attribute, but we
+	 * do not use that feature in this routine.)
 	 */
+	cleanLength = cleanTupType->natts;
 	if (cleanLength > 0)
 	{
 		cleanMap = (AttrNumber *) palloc(cleanLength * sizeof(AttrNumber));
@@ -191,42 +93,12 @@ ExecInitJunkFilter(List *targetList, TupleDesc tupType,
 		foreach(t, targetList)
 		{
 			TargetEntry *tle = lfirst(t);
+			Resdom	   *resdom = tle->resdom;
 
-			if (tle->resdom != NULL)
+			if (!resdom->resjunk)
 			{
-				resdom = tle->resdom;
-				expr = tle->expr;
-				resjunk = resdom->resjunk;
-				if (!resjunk)
-				{
-					cleanMap[cleanResno - 1] = resdom->resno;
-					cleanResno++;
-				}
-			}
-			else
-			{
-#ifdef SETS_FIXED
-				List		fjListP;
-				List		fjList = lfirst(t);
-				Fjoin		fjNode = (Fjoin) lfirst(fjList);
-
-				/* what the hell is this????? */
-				resdom = (Resdom) lfirst(get_fj_innerNode(fjNode));
-#endif
-
-				cleanMap[cleanResno - 1] = tle->resdom->resno;
+				cleanMap[cleanResno - 1] = resdom->resno;
 				cleanResno++;
-
-#ifdef SETS_FIXED
-				foreach(fjListP, lnext(fjList))
-				{
-					TargetEntry *tle = lfirst(fjListP);
-
-					resdom = tle->resdom;
-					cleanMap[cleanResno - 1] = resdom->resno;
-					cleanResno++;
-				}
-#endif
 			}
 		}
 	}
@@ -239,47 +111,96 @@ ExecInitJunkFilter(List *targetList, TupleDesc tupType,
 	junkfilter = makeNode(JunkFilter);
 
 	junkfilter->jf_targetList = targetList;
-	junkfilter->jf_length = len;
-	junkfilter->jf_tupType = tupType;
-	junkfilter->jf_cleanTargetList = cleanTargetList;
-	junkfilter->jf_cleanLength = cleanLength;
 	junkfilter->jf_cleanTupType = cleanTupType;
 	junkfilter->jf_cleanMap = cleanMap;
-	junkfilter->jf_junkContext = junkContext;
 	junkfilter->jf_resultSlot = slot;
 
 	if (slot)
 		ExecSetSlotDescriptor(slot, cleanTupType, false);
 
-	MemoryContextSwitchTo(oldContext);
+	return junkfilter;
+}
+
+/*
+ * ExecInitJunkFilterConversion
+ *
+ * Initialize a JunkFilter for rowtype conversions.
+ *
+ * Here, we are given the target "clean" tuple descriptor rather than
+ * inferring it from the targetlist.  The target descriptor can contain
+ * deleted columns.  It is assumed that the caller has checked that the
+ * non-deleted columns match up with the non-junk columns of the targetlist.
+ */
+JunkFilter *
+ExecInitJunkFilterConversion(List *targetList,
+							 TupleDesc cleanTupType,
+							 TupleTableSlot *slot)
+{
+	JunkFilter *junkfilter;
+	int			cleanLength;
+	AttrNumber *cleanMap;
+	ListCell   *t;
+	int			i;
+
+	/*
+	 * Calculate the mapping between the original tuple's attributes and
+	 * the "clean" tuple's attributes.
+	 *
+	 * The "map" is an array of "cleanLength" attribute numbers, i.e. one
+	 * entry for every attribute of the "clean" tuple. The value of this
+	 * entry is the attribute number of the corresponding attribute of the
+	 * "original" tuple.  We store zero for any deleted attributes, marking
+	 * that a NULL is needed in the output tuple.
+	 */
+	cleanLength = cleanTupType->natts;
+	if (cleanLength > 0)
+	{
+		cleanMap = (AttrNumber *) palloc0(cleanLength * sizeof(AttrNumber));
+		t = list_head(targetList);
+		for (i = 0; i < cleanLength; i++)
+		{
+			if (cleanTupType->attrs[i]->attisdropped)
+				continue;		/* map entry is already zero */
+			for (;;)
+			{
+				TargetEntry *tle = lfirst(t);
+				Resdom	   *resdom = tle->resdom;
+
+				t = lnext(t);
+				if (!resdom->resjunk)
+				{
+					cleanMap[i] = resdom->resno;
+					break;
+				}
+			}
+		}
+	}
+	else
+		cleanMap = NULL;
+
+	/*
+	 * Finally create and initialize the JunkFilter struct.
+	 */
+	junkfilter = makeNode(JunkFilter);
+
+	junkfilter->jf_targetList = targetList;
+	junkfilter->jf_cleanTupType = cleanTupType;
+	junkfilter->jf_cleanMap = cleanMap;
+	junkfilter->jf_resultSlot = slot;
+
+	if (slot)
+		ExecSetSlotDescriptor(slot, cleanTupType, false);
 
 	return junkfilter;
 }
 
-/*-------------------------------------------------------------------------
- * ExecFreeJunkFilter
- *
- * Release the data structures created by ExecInitJunkFilter.
- *-------------------------------------------------------------------------
- */
-void
-ExecFreeJunkFilter(JunkFilter *junkfilter)
-{
-	/*
-	 * Since the junkfilter is inside its own context, we just have to
-	 * delete the context and we're set.
-	 */
-	MemoryContextDelete(junkfilter->jf_junkContext);
-}
-
-/*-------------------------------------------------------------------------
+/*
  * ExecGetJunkAttribute
  *
  * Given a tuple (slot), the junk filter and a junk attribute's name,
  * extract & return the value and isNull flag of this attribute.
  *
  * It returns false iff no junk attribute with such name was found.
- *-------------------------------------------------------------------------
  */
 bool
 ExecGetJunkAttribute(JunkFilter *junkfilter,
@@ -289,11 +210,8 @@ ExecGetJunkAttribute(JunkFilter *junkfilter,
 					 bool *isNull)
 {
 	List	   *targetList;
-	List	   *t;
-	Resdom	   *resdom;
+	ListCell   *t;
 	AttrNumber	resno;
-	char	   *resname;
-	bool		resjunk;
 	TupleDesc	tupType;
 	HeapTuple	tuple;
 
@@ -307,11 +225,10 @@ ExecGetJunkAttribute(JunkFilter *junkfilter,
 	foreach(t, targetList)
 	{
 		TargetEntry *tle = lfirst(t);
+		Resdom	   *resdom = tle->resdom;
 
-		resdom = tle->resdom;
-		resname = resdom->resname;
-		resjunk = resdom->resjunk;
-		if (resjunk && (strcmp(resname, attrName) == 0))
+		if (resdom->resjunk && resdom->resname &&
+			(strcmp(resdom->resname, attrName) == 0))
 		{
 			/* We found it ! */
 			resno = resdom->resno;
@@ -329,14 +246,14 @@ ExecGetJunkAttribute(JunkFilter *junkfilter,
 	 * Now extract the attribute value from the tuple.
 	 */
 	tuple = slot->val;
-	tupType = junkfilter->jf_tupType;
+	tupType = slot->ttc_tupleDescriptor;
 
 	*value = heap_getattr(tuple, resno, tupType, isNull);
 
 	return true;
 }
 
-/*-------------------------------------------------------------------------
+/*
  * ExecRemoveJunk
  *
  * Construct and return a tuple with all the junk attributes removed.
@@ -344,49 +261,48 @@ ExecGetJunkAttribute(JunkFilter *junkfilter,
  * Note: for historical reasons, this does not store the constructed
  * tuple into the junkfilter's resultSlot.  The caller should do that
  * if it wants to.
- *-------------------------------------------------------------------------
  */
 HeapTuple
 ExecRemoveJunk(JunkFilter *junkfilter, TupleTableSlot *slot)
 {
+#define PREALLOC_SIZE	64
 	HeapTuple	tuple;
 	HeapTuple	cleanTuple;
 	AttrNumber *cleanMap;
 	TupleDesc	cleanTupType;
 	TupleDesc	tupType;
 	int			cleanLength;
-	bool		isNull;
+	int			oldLength;
 	int			i;
 	Datum	   *values;
 	char	   *nulls;
-	Datum		values_array[64];
-	char		nulls_array[64];
+	Datum	   *old_values;
+	char	   *old_nulls;
+	Datum		values_array[PREALLOC_SIZE];
+	Datum		old_values_array[PREALLOC_SIZE];
+	char		nulls_array[PREALLOC_SIZE];
+	char		old_nulls_array[PREALLOC_SIZE];
 
 	/*
 	 * get info from the slot and the junk filter
 	 */
 	tuple = slot->val;
+	tupType = slot->ttc_tupleDescriptor;
+	oldLength = tupType->natts + 1;			/* +1 for NULL */
 
-	tupType = junkfilter->jf_tupType;
 	cleanTupType = junkfilter->jf_cleanTupType;
-	cleanLength = junkfilter->jf_cleanLength;
+	cleanLength = cleanTupType->natts;
 	cleanMap = junkfilter->jf_cleanMap;
 
-	/*
-	 * Handle the trivial case first.
-	 */
-	if (cleanLength == 0)
-		return (HeapTuple) NULL;
-
 	/*
 	 * Create the arrays that will hold the attribute values and the null
-	 * information for the new "clean" tuple.
+	 * information for the old tuple and new "clean" tuple.
 	 *
 	 * Note: we use memory on the stack to optimize things when we are
 	 * dealing with a small number of attributes. for large tuples we just
 	 * use palloc.
 	 */
-	if (cleanLength > 64)
+	if (cleanLength > PREALLOC_SIZE)
 	{
 		values = (Datum *) palloc(cleanLength * sizeof(Datum));
 		nulls = (char *) palloc(cleanLength * sizeof(char));
@@ -396,36 +312,57 @@ ExecRemoveJunk(JunkFilter *junkfilter, TupleTableSlot *slot)
 		values = values_array;
 		nulls = nulls_array;
 	}
+	if (oldLength > PREALLOC_SIZE)
+	{
+		old_values = (Datum *) palloc(oldLength * sizeof(Datum));
+		old_nulls = (char *) palloc(oldLength * sizeof(char));
+	}
+	else
+	{
+		old_values = old_values_array;
+		old_nulls = old_nulls_array;
+	}
 
 	/*
-	 * Exctract one by one all the values of the "clean" tuple.
+	 * Extract all the values of the old tuple, offsetting the arrays
+	 * so that old_values[0] is NULL and old_values[1] is the first
+	 * source attribute; this exactly matches the numbering convention
+	 * in cleanMap.
+	 */
+	heap_deformtuple(tuple, tupType, old_values + 1, old_nulls + 1);
+	old_values[0] = (Datum) 0;
+	old_nulls[0] = 'n';
+
+	/*
+	 * Transpose into proper fields of the new tuple.
 	 */
 	for (i = 0; i < cleanLength; i++)
 	{
-		values[i] = heap_getattr(tuple, cleanMap[i], tupType, &isNull);
+		int			j = cleanMap[i];
 
-		if (isNull)
-			nulls[i] = 'n';
-		else
-			nulls[i] = ' ';
+		values[i] = old_values[j];
+		nulls[i] = old_nulls[j];
 	}
 
 	/*
 	 * Now form the new tuple.
 	 */
-	cleanTuple = heap_formtuple(cleanTupType,
-								values,
-								nulls);
+	cleanTuple = heap_formtuple(cleanTupType, values, nulls);
 
 	/*
 	 * We are done.  Free any space allocated for 'values' and 'nulls' and
 	 * return the new tuple.
 	 */
-	if (cleanLength > 64)
+	if (values != values_array)
 	{
 		pfree(values);
 		pfree(nulls);
 	}
+	if (old_values != old_values_array)
+	{
+		pfree(old_values);
+		pfree(old_nulls);
+	}
 
 	return cleanTuple;
 }
